#include "./samc.h"

/*************** For RET_COST structure ******************/

/* Function to initialize the structure */
RET_COST * cost_init(nrow, ngrp, ndim)
int nrow, ngrp, ndim;
{
  RET_COST *rcost;

  rcost = (RET_COST *) malloc(sizeof(RET_COST));
  CHECK_MEM(rcost);
  if (nrow) {
    rcost->Yhat = dvector(1,nrow);
  } else {
    rcost->Yhat = NULL;
  }
  if (ngrp) {
    rcost->groups = ivector(1,ngrp);
  } else {
    rcost->groups = NULL;
  }
  if (ndim) {
    rcost->col_ids = ivector(1,ndim);
    rcost->beta    = dvector(1,ndim);
  } else {
    rcost->col_ids = NULL;
    rcost->beta    = NULL;
  }

  rcost->num_grp   = ngrp;
  rcost->dim       = ndim;
  rcost->converged = 0;
  rcost->value     = -1.0;
  rcost->deviance  = -1.0;

  return(rcost);

} /* END: cost_init */

/* Function to free the structure */
void cost_free(rcost, nrow)
RET_COST *rcost;
int nrow;
{
  if (rcost->Yhat) free_dvector(rcost->Yhat,1,nrow);
  if (rcost->groups) free_ivector(rcost->groups,1,rcost->num_grp);
  if (rcost->col_ids) free_ivector(rcost->col_ids,1,rcost->dim);
  if (rcost->beta) free_dvector(rcost->beta,1,rcost->dim);

  free(rcost);

} /* END: cost_free */

/* Function to free all models */
void models_free(parms)
PARMS *parms;
{
  int i, DataNum=parms->DataNum;
  RET_COST **Models=parms->Models;


  for (i=1; i<=parms->Model_n; i++) cost_free(Models[i], DataNum);
  free(Models);

} /* END: cost_free */

/* Function to create a RET_COST structure and copy */
RET_COST * cost_create(obj, nrow)
RET_COST *obj;
int nrow;  /* Use 0 to not allocate and copy the Yhat */
{
  RET_COST *rcost;
  double *p1, *p2;
  int i, n, *ptri1, *ptri2;

  rcost = cost_init(nrow, obj->num_grp, obj->dim);

  rcost->value     = obj->value;
  rcost->deviance  = obj->deviance;
  rcost->dim       = obj->dim;
  rcost->num_grp   = obj->num_grp;
  rcost->converged = obj->converged;

  p1 = rcost->Yhat;
  p2 = obj->Yhat;
  if (nrow) {
    for (i=1; i<=nrow; i++) p1[i] = p2[i];
  }

  /* Copy groups, column ids, and beta */
  n = obj->num_grp;
  if (n) {
    for (i=1, ptri1=rcost->groups, ptri2=obj->groups; i<=n; i++) ptri1[i] = ptri2[i];
  }
  n = obj->dim;
  if (n) {
    for (i=1, ptri1=rcost->col_ids, ptri2=obj->col_ids; i<=n; i++) ptri1[i] = ptri2[i];
    for (i=1, p1=rcost->beta, p2=obj->beta; i<=n; i++) p1[i] = p2[i];
  }

  return(rcost);

} /* END: cost_create */

/* Function to copy a RET_COST structure. Vectors in ret must have sufficient length */
void cost_copy(obj, nrow, ret)
RET_COST *obj, *ret;
int nrow;  /* Use 0 to not allocate and copy the Yhat */
{
  double *p1, *p2;
  int i, n, *ptri1, *ptri2;

  ret->value     = obj->value;
  ret->deviance  = obj->deviance;
  ret->dim       = obj->dim;
  ret->num_grp   = obj->num_grp;
  ret->converged = obj->converged;

  if (nrow) {
    p1 = ret->Yhat;
    p2 = obj->Yhat;
    for (i=1; i<=nrow; i++) p1[i] = p2[i];
  }

  /* Copy groups, column ids, and beta */
  n = obj->num_grp;
  if (n) {
    for (i=1, ptri1=ret->groups, ptri2=obj->groups; i<=n; i++) ptri1[i] = ptri2[i];
  }
  n = obj->dim;
  if (n) {
    for (i=1, ptri1=ret->col_ids, ptri2=obj->col_ids; i<=n; i++) ptri1[i] = ptri2[i];
    for (i=1, p1=ret->beta, p2=obj->beta; i<=n; i++) p1[i] = p2[i];
  }

} /* END: cost_copy */


/* Function to print a RET_COST structure */
void print_cost(rcost)
RET_COST *rcost;
{
  printf("Yhat=%p, value=%g, deviance=%g, dim=%d, num_grp=%d, converged=%d\n", rcost->Yhat, rcost->value, rcost->deviance, rcost->dim,\
            rcost->num_grp, rcost->converged);
  printf("groups=%p, col_ids=%p, beta=%p\n", rcost->groups, rcost->col_ids, rcost->beta);
  print_iVec(rcost->groups, rcost->num_grp, "groups");
  print_iVec(rcost->col_ids, rcost->dim, "col_ids");
  print_dVec(rcost->beta, rcost->dim, "beta");

} 
/* END: print_cost */

/* Function to print the Models vector (for debugging) */
void print_Models(which, parms)
int which;
PARMS *parms;
{
  int i, Model_n=parms->Model_n;
  RET_COST *rcost, **Models=parms->Models;

  for (i=1; i<=Model_n; i++) {
    if (which > 0) {
      rcost = Models[which];
      i = which;
    } else {
      rcost = Models[i];
    }
    printf(" ***Model %d ***\n", i);  
    print_cost(rcost);
    printf("\n");
    if (which > 0) break;
  }

} /* END: print_Models */

/* Function to write a single model to an open file */
void write_model(fpt, r, resample)
FILE *fpt;
RET_COST *r;
int resample;  /* 0 or 1 if called from resample */
{
  int n, j, *ptri;
  double *ptrd;

  if (!resample) {
    fprintf(fpt, "%g %g %d %d %d\n", r->value, r->deviance, r->dim, r->num_grp, r->converged);
    n = r->num_grp;
    if (n) {
      ptri = r->groups;
      for (j=1; j<n; j++) fprintf(fpt, "%d ", ptri[j]);
      fprintf(fpt, "%d\n", ptri[n]);
    }
  } else {
    fprintf(fpt, "%d\n", r->dim);
  }
  n = r->dim;
  if (n) {
    ptri = r->col_ids;
    for (j=1; j<n; j++) fprintf(fpt, "%d ", ptri[j]);
    fprintf(fpt, "%d\n", ptri[n]);

    ptrd = r->beta;
    for (j=1; j<n; j++) fprintf(fpt, "%g ", ptrd[j]);
    fprintf(fpt, "%g\n", ptrd[n]);
  }

} /* END: write_model */

/* Funtion to write Models to a file */
void write_Models(parmfile, parms)
char *parmfile;
PARMS *parms;
{
  char *tempstr=NULL;
  FILE *fpt;
  int i, j, *ptri, n, Model_n=parms->Model_n;
  RET_COST *r, **Models=parms->Models;
  double *ptrd;

  if (Model_n < 1) return;

  tempstr = (char *) malloc(MAXFILELEN*sizeof(char));
  CHECK_MEM(tempstr);

  i = get_option(parmfile, "out_models", tempstr, 0);
  if (i) {
    free(tempstr);
    return;
  }
  
  fpt = open_file(tempstr, "w");
  free(tempstr);
  
  fprintf(fpt, "%d\n", Model_n);
  for (i=1; i<=Model_n; i++) {
    r = Models[i];
    write_model(fpt, r, 0);

  /*
    fprintf(fpt, "%g %g %d %d %d\n", r->value, r->deviance, r->dim, r->num_grp, r->converged);
    n = r->num_grp;
    if (n) {
      ptri = r->groups;
      for (j=1; j<n; j++) fprintf(fpt, "%d ", ptri[j]);
      fprintf(fpt, "%d\n", ptri[n]);
    }
    n = r->dim;
    if (n) {
      ptri = r->col_ids;
      for (j=1; j<n; j++) fprintf(fpt, "%d ", ptri[j]);
      fprintf(fpt, "%d\n", ptri[n]);

      ptrd = r->beta;
      for (j=1; j<n; j++) fprintf(fpt, "%g ", ptrd[j]);
      fprintf(fpt, "%g\n", ptrd[n]);
    }
  */
  }
  fclose(fpt);

} /* END: write_Models */ 

/* Function to read in file of saved models */
void set_Models(parmfile, parms)
char *parmfile;
PARMS *parms;
{
  char *tempstr;
  FILE *fpt;
  int i, j, *ptri, n, dim, ngrp, conv;
  RET_COST *r, **Models;
  double *ptrd, value, deviance;
  float x;

  tempstr = (char *) malloc(MAXFILELEN*sizeof(char));
  CHECK_MEM(tempstr);
  
  i = get_option(parmfile, "in_models", tempstr, 0);
  if (i) {
    /* Input file not specified */
    free(tempstr);
    parms->Model_n = 0;
    parms->Model_maxn = parms->Model_maxn;
    parms->Models = (RET_COST **) malloc((parms->Model_maxn+1)*sizeof(RET_COST *));
    CHECK_MEM(parms->Models);
    return;
  }

  /* Input file was specified */
  fpt = open_file(tempstr, "r");
  free(tempstr);

  /* Get the number of models */
  fscanf(fpt, "%d", &n);
  parms->Model_n = n;
  parms->Model_maxn = MAX(n, parms->Model_maxn);
  parms->Models = (RET_COST **) malloc((parms->Model_maxn+1)*sizeof(RET_COST *));
  CHECK_MEM(parms->Models);
  Models = parms->Models;

  /* Read in each model */
  for (i=1; i<=parms->Model_n; i++) { 

    /* Read in value, deviance, dim, num_grp, converged */
    fscanf(fpt, "%g", &x);
    value = x;
    fscanf(fpt, "%g", &x);
    deviance = x;
    fscanf(fpt, "%d", &n);
    dim = n;
    fscanf(fpt, "%d", &n);
    ngrp = n;
    fscanf(fpt, "%d", &n);
    conv = n;

    /* Create a new structure */
    r = cost_init(0, ngrp, dim);
    r->value     = value;
    r->deviance  = deviance;
    r->dim       = dim;
    r->num_grp   = ngrp;
    r->converged = conv;

    /* Read in groups if num_grp > 0 */
    if (r->num_grp) {
      ptri = r->groups;
      for (j=1; j<=r->num_grp; j++) {
        fscanf(fpt, "%d", &n);
        ptri[j] = n;
      }
    }

    /* Read in col_ids and beta if dim > 0 */
    if (r->dim) {
      ptri = r->col_ids;
      for (j=1; j<=r->dim; j++) {
        fscanf(fpt, "%d", &n);
        ptri[j] = n;
      }
      ptrd = r->beta;
      for (j=1; j<=r->dim; j++) {
        fscanf(fpt, "%g", &x);
        ptrd[j] = x;
      }
    }

    Models[i] = r;
  }
  fclose(fpt);

} /* END: set_Models */


/*****************************************************************************/

/*****************************************************************************
***************** String functions *******************************************
******************************************************************************/

/* Function to get the next token */
char * next_token(str, n, sep)
char *str, sep;
int n;
{
  int i;
  char *pc, c1;

  if (str == NULL) return(NULL);
  
  /* Loop till a sep character or end of string */
  for (i=0, pc=str; i<n; i++, pc++) {
    c1 = *pc;
    if (c1 == sep) {
      pc++;
      return(pc);
    } else if (c1 == END_OF_STR) {
      return(NULL);
    }
  }
  return(NULL);

} /* END: next_token */

/* Function to copy a token from a line read in to a field and adding '\0' */
int copy_token(ret, str, sep, n)
char *ret; /* Return string */
char *str; /* Must start at the correct position */
char sep;
int n;
{
  int i, flag=2;
  char *pc, *pret, c1;
  
  /* The end-of-line character should appear before the end-of-string character */
 
  /*n = MIN(n, strlen(str));*/
  for (i=0, pc=str, pret=ret; i<n; i++, pc++, pret++) {
    c1 = *pc;

    if ((c1 != sep) && (c1 != END_OF_LINE) && (c1 != CAR_RETURN) && (c1 != EOF)) {
      *pret = *pc;
      flag = 1;
    } else {
       *pret = END_OF_STR;
       flag--;
       break;
    }
  }

  return(flag);

} /* END: copy_token */

/* Function to count a char in a string */
int count_char(str, c1)
char *str, c1;
{
  int i, len, n;

  len = strlen(str);
  n = 0;
  for (i=0; i<len; i++) {
    if (str[i] == c1) n++;
  }

  return(n);

} /* END: count_char */

/* Function to return an before a delimiter */
int getIntBeforeSep(str, sep)
char *str, sep;
{
  char cnum[101], c1, *end;
  int num, i, flag=0;

  for (i=0; 1<100; i++) {
    c1 = str[i];  
    if (c1 == sep) {
      flag = 1;
      cnum[i] = END_OF_STR;
      num = (int) strtol(cnum, &end, 10);
      return(num);
    } else if ((c1 == ',') || (c1 == END_OF_STR) || (c1 == END_OF_LINE) || (c1 == CAR_RETURN)) {
      return(NOTFOUND);
    } else {
      cnum[i] = c1;
    }
  }

  return(NOTFOUND);

} /* getIntBeforeSep */

/* Function to get the integers from a string such as "1-7,13,20-25" */
int getIntsFromStr(str, intLen, n, retVec)
char *str;
int intLen; /* Maximum number of characters for one of the ints */
int n;  /* Length of retVec (1 + max number) */
int *retVec;
{
  int i, strIndex, len, index, start, num, rangeFlag, im1;
  char *end, chr[intLen], c1;

  /* Initialize */
  for (i=0; i<n; i++) retVec[i] = 0;
  len = (int) strlen(str);
  
  strIndex = 0;
  start = 0;
  rangeFlag = 0;
  index = 0;
  c1 = 'e';
  im1 = intLen - 1;
  while(1) {
    if (strIndex <= len) c1 = str[strIndex];

    if ((strIndex > len) || (c1 == ',')) {

      if (!index) {
        printf("\nERROR1: The list %s is not valid\n", str);
        exit(1);
      }

      /* Get number */
      chr[index] = END_OF_STR;
      num = (int) strtol(chr, &end, 10);
      if ((num < 1) || (num > n)) {
        printf("\nERROR2: The list %s is not valid\n", str);
        exit(1);
      }

      if (rangeFlag) {
        if ((!start) || (num < start)) {
          printf("\nERROR3: The list %s is not valid\n", str);
          exit(1);
        }
        for (i=start; i<=num; i++) retVec[i] = 1;
      } else {
        retVec[num] = 1;
      }

      rangeFlag = 0;
      index = 0;
      start = 0;

      if (strIndex > len) break;
    } else {
      if (c1 == '-') {
      
        /* Get previous number */
        chr[index] = END_OF_STR;
        start = (int) strtol(chr, &end, 10);
        if ((start < 1) || (start > n)) {
          printf("\nERROR4: The list %s is not valid\n", str);
          exit(1);
        }

        rangeFlag = 1;
        index = 0;
      
      } else {
        if (index > im1) {
          printf("\nERROR5: The list %s is not valid\n", str);
          exit(1);
        }
        chr[index] = c1;
        index++;
      }

    }
    strIndex++;
  }

  /* Get the maximum chr number */
  for (i=n-1; i>-1; i--) {
    if (retVec[i]) return(i);
  }

  return(0);

} /* END: getIntsFromStr */ 

/*****************************************************************************
***************** Reading files **********************************************
******************************************************************************/

/* Function to read the group file. Col 1 is the group number, col 2 is the comma seperated list of variable numbers */
int read_groups(datafile, alwaysInModel, maxCol, X)
char *datafile;
int *alwaysInModel, maxCol;
int **X;
{
  FILE *fpt;
  int i, j, *used, *cols, header;
  char line[BUFFERSIZE], sep, *token, id[51], *end;
  size_t size;
  int numCols, nc, ngroup, flag, group, group0=0;
  
 
  nc = maxCol + 1;
  used = (int *) malloc(nc*sizeof(int));
  CHECK_MEM(used);
  cols = (int *) malloc(nc*sizeof(int));
  CHECK_MEM(cols);

  for (i=0; i<nc; i++) used[i] = 0;
  for (i=1; i<=maxCol; i++) {
    for (j=1; j<=maxCol; j++) X[i][j] = 0;
  }
  

  /* Get the file delimiter */
  sep = getDelimiter(datafile, BUFFERSIZE);
  header = getHeader(datafile, sep, 2, BUFFERSIZE);

  ngroup = 0;
  size = sizeof(line);
  fpt = open_file(datafile, "r");
 
  while (fgets(line, size, fpt) != NULL) {   
    if (header) {
      header = 0;
      continue;
    }
    
    /* Get the group number */
    flag = copy_token(id, line, sep, 50);
    if (flag) {
      printf("ERROR with group id\n");
      exit(1);
    }
    group = (int) strtol(id, &end, 10);
    if (group != group0 + 1) {
      printf("ERROR: group id values must be 1, 2 ...\n");
      exit(1);
    }
    group0 = group;
    token = next_token(line, 1000, sep);

    i = getIntsFromStr(token, 6, maxCol, used);

    /* Get the numbers in the list */
    numCols = 0;
    for (i=1; i<=maxCol; i++) {
      if ((used[i]) && (alwaysInModel[i])) {
        printf("ERROR: column %i is always in the model. Groups refer to variables not always in the model.\n", i);
        exit(1);
      }
      if (used[i]) {
        cols[numCols] = i;
        numCols++; 
      }
    }

    /* All variables in list are grouped together */
    for (i=0; i<numCols; i++) X[group][cols[i]] = 1;
    
    ngroup++;
  }
  fclose(fpt);

  /* Get the total number of groups by adding in the columns in their own group */
  /*
  for (i=1; i<=maxCol; i++) {
    if (!alwaysInModel[i]) {
      k = 0;
      for (j=1; j<=maxCol; j++) {
        if (X[i][j]) {
          k = 1;
          break; 
        }
      }
      if (!k) {
        X[i][1] = i;
        ngroup++;
      }
    }
  }
  */

  free(used);
  free(cols);
  
  return(ngroup);

} /* END: read_groups */

/* Function to read the data file and put it into a matrix */
void read_data(datafile, sep, mat, nr, nc, bufferSize, header)
char *datafile;
char sep; /* File delimiter */
double **mat; /* Defined from 1 to nr, 1 to nc */
int nr, nc;
long bufferSize;
int header; /* 0 or 1 for a header of variable names */
{
  FILE *fpt;
  int i, j, k;
  char line[bufferSize], *token, temp[101], *end;
  size_t size;
  
  size = sizeof(line);
  fpt = open_file(datafile, "r");
  if (header) fgets(line, size, fpt);
   
  for (i=1; i<=nr; i++) {
    if (fgets(line, size, fpt) == NULL) {
      printf("End of file reached before %d rows read in\n", nr);
      exit(1);
    }
    token = line;
    for (j=1; j<=nc; j++) {
      k = copy_token(temp, token, sep, 100);
      if (k) {
        printf("copy ERROR on line %d, column %d of file %s\n", i, j, datafile);
        exit(1);
      }
      mat[i][j] = strtod(temp, &end);
      if (j < nc) {
        token = next_token(token, 100, sep);
        if (token == NULL) {
          printf("next ERROR on line %d, column %d of file %s\n", i, j, datafile);
          exit(1);
        }
      }
    }

  }

  fclose(fpt);

} /* END: read_data */

/* Function to read the data file and put it into an integer matrix */
void read_idata(datafile, sep, mat, nr, nc, bufferSize, header)
char *datafile;
char sep; /* File delimiter */
int **mat; /* Defined from 1 to nr, 1 to nc */
int nr, nc;
long bufferSize;
int header; /* 0 or 1 for a header of variable names */
{
  FILE *fpt;
  int i, j, k;
  char line[bufferSize], *token, temp[101], *end;
  size_t size;
  
  size = sizeof(line);
  fpt = open_file(datafile, "r");
  if (header) fgets(line, size, fpt);
  for (i=1; i<=nr; i++) {
    if (fgets(line, size, fpt) == NULL) {
      printf("End of file reached before %d rows read in\n", nr);
      exit(1);
    }
    token = line;
    for (j=1; j<=nc; j++) {
      k = copy_token(temp, token, sep, 100);
      if (k) {
        printf("copy ERROR on line %d, column %d of file %s\n", i, j, datafile);
        exit(1);
      }
      mat[i][j] = (int) strtol(temp, &end, 10);
      if (j < nc) {
        token = next_token(token, 100, sep);
        if (token == NULL) {
          printf("next ERROR on line %d, column %d of file %s\n", i, j, datafile);
          exit(1);
        }
      }
    }

  }

  fclose(fpt);

} /* END: read_idata */

/* Function to read the interactions data file */
int ** read_interactions(datafile, ngrp, retVec)
char *datafile;
int ngrp; /* Number of groups */
int *retVec;  /* Vector of length # of groups, starting at index 1 for the number of groups in interaction i*/
{
  FILE *fpt;
  int i, j, k, header, *grpids, intGrp, nint, **mat, row=0, id;
  char line[BUFFERSIZE_INT], *token, temp[101], *end, sep;
  size_t size;
  
  sep = getDelimiter(datafile, BUFFERSIZE_INT);
  i   = getNcols(datafile, sep, BUFFERSIZE_INT);
  header = getHeader(datafile, sep, i, BUFFERSIZE_INT);

  grpids = ivector(1, ngrp);
  mat = (int **) malloc((ngrp+1)*sizeof(int *));
  CHECK_MEM(mat);
  for (i=0; i<=ngrp; i++) mat[i] = NULL;
  for (i=1; i<=ngrp; i++) retVec[i] = 0;

  size = sizeof(line);
  fpt = open_file(datafile, "r");
  if (header) {
    fgets(line, size, fpt);
    row++;
  }
  while (fgets(line, size, fpt) != NULL) {
    row++;
    token = line;
    nint  = 0;
    for (j=0; j<ngrp; j++) {
      k = copy_token(temp, token, sep, 100);
      if (k) {
        printf("copy ERROR on line %d, column %d of file %s\n", row, j+1, datafile);
        exit(1);
      }
      id = (int) strtol(temp, &end, 10);
      if ((id < 1) || (id > ngrp)) {
        printf("ERROR on line %d, column %d of file %s\n", row, j+1, datafile);
        exit(1);
      }
      if (!j) {
        /* First one is the interaction group number */
        intGrp = id;
        if (retVec[intGrp]) {
          printf("ERROR: With interaction file. Interaction %d cannot be defined more than once\n", intGrp);
          exit(1);
        }
      } else {
        nint++;
        grpids[nint] = id; 
      }
      token = next_token(token, 100, sep);
      if (token == NULL) break;
    }
    if (nint >= ngrp) {
      printf("ERROR: Too many groups defined for interaction group %d\n", intGrp);
      exit(1);
    }
    if (!nint) {
      printf("ERROR on line %d of file %s\n", row, datafile);
      exit(1);
    }
 
    mat[intGrp] = (int *) malloc(nint*sizeof(int));
    CHECK_MEM(mat[intGrp]);

    /* Store the group ids in mat */
    for (j=0; j<nint; j++) mat[intGrp][j] = grpids[j+1];
    retVec[intGrp] = nint;
  }

  fclose(fpt);
  free_ivector(grpids, 1, ngrp);

  return(mat);

} /* END: read_interactions */



/* Function for opening a data set in read mode */
FILE * open_file(datafile, mode)
char *datafile, *mode;
{
  FILE *fpt=0;

  fpt = fopen(datafile, mode);
  if (fpt == 0) {
    printf("\n Error opening file: %s \n", datafile);
    exit(0);
  }

  return(fpt);
}

/* Function for getting the number of columns in a delimited data file */
int getNcols(datafile, sep, bufferSize)
char *datafile, sep;
long bufferSize;
{
  FILE *fpt;
  int i ;
  char line[bufferSize];

  fpt = open_file(datafile, "r");
  fgets(line, sizeof(line), fpt);
  fclose(fpt);
  i = count_char(line, sep) + 1;

  return(i);
}

/* Function to determine the file delimiter */
char getDelimiter(datafile, bufferSize)
char *datafile;
long bufferSize;
{
  FILE *fpt;
  int i, len, n_tab=0, n_comma=0, n_space=0;
  char line[bufferSize], *pc, c1;

  fpt = open_file(datafile, "r");
  fgets(line, sizeof(line), fpt);
  fclose(fpt);

  len = (int) strlen(line);
  for (i=0, pc=line; i<len; i++, pc++) {
    c1 = *pc;
    if (c1 == '\t') {
      n_tab++;
    } else if (c1 == ' ') {
      n_space++;
    } else if (c1 == ',') {
      n_comma++;
    } 
  }

  if (n_tab) {
    return('\t');
  } else if (n_comma) {
    return(',');
  } else if (n_space) {
    return(' ');
  } else {
    return('\t');
  }

} /* getDelimiter */

/* Function to determine if a file has a header of variable names */
int getHeader(datafile, sep, ncols, bufferSize)
char *datafile, sep;
int ncols;
long bufferSize;
{
  FILE *fpt;
  int i, ret=0;
  char line[bufferSize], c1, *token, *end, cnum[2]; 

  fpt = open_file(datafile, "r");
  fgets(line, sizeof(line), fpt);
  fclose(fpt);

  cnum[1] = END_OF_STR;
  c1 = line[0];
  cnum[0] = c1;
  if ((c1 != '0') && (c1 != '.') && (!strtol(cnum, &end, 10))) {
    ret = 1;
  } else {
    return(0);
  }
  token = line;
  for (i=1; i<ncols; i++) {
    token = next_token(token, bufferSize, sep);
    if (token == NULL) return(ret);
    c1 = token[0];
    cnum[0] = c1;
    if ((c1 != '0') && (c1 != '.') && (!strtol(cnum, &end, 10))) {
      ret = 1;
    } else {
      return(0);
    }
  }

  return(ret);

} /* END: getHeader */

/* Function for getting the number of rows in a data file */
int getNrows(datafile, bufferSize)
char *datafile;
long bufferSize;
{
  FILE *fpt;
  int i = 0;
  char line[bufferSize];
  size_t size;

  size = sizeof(line);
  fpt = open_file(datafile, "r");
  while(fgets(line, size, fpt) != NULL) i++;
  fclose(fpt);
  return(i);
}

/* Function for reading an option from a file */
int get_option(parmfile, option, outstring, errorFlag)
char *parmfile, *option, *outstring;
int errorFlag;
{
  FILE *fpt;
  char value[500]="", name[500];
  int temp, len, flag=0, f1, f2, val1_flag;
  
  len = (int) strlen(option);

  fpt = open_file(parmfile, "r");

  /* Determine the order from row 1*/
  temp = fscanf(fpt, "%s %s", value, name);
  f1 = strncmp("VALUE", value, 5);
  f2 = strncmp("VALUE", name, 5);
  if ((!f1) && (!f2)) {
    printf("\n ERROR: Row 1 of input file must have the column names VALUE and NAME \n");
    fclose(fpt);
    exit(1);
  }
  if (!f1) {
    val1_flag = 1;
  } else {
    val1_flag = 0;
  }

  while(!feof(fpt)) 
  {   
    if (val1_flag) {
      temp = fscanf(fpt, "%s %s", value, name);
    } else {
      temp = fscanf(fpt, "%s %s", name, value);
    }
    if (temp == EOF) break;
    temp = strncmp(name, option, len);
    if (!temp) {
      flag = 1;
      break;
    }
  }
  fclose(fpt);

  if (!flag) {
    if (errorFlag) {
      printf("\n ERROR: The following option name was not found in the input file: \n");
      printf("%s\n", option);
      exit(1);
    } else {
      return(1);
    }
  }

  len = (int) strlen(value);
  strncpy(outstring, value, len);
  outstring[len] = '\0';

  return(0);
} 

void read_parms(datastr, parmfile)
PARMS *datastr;
char *parmfile;
{
  FILE *fpt;
  char value[MAXFILELEN];
  float x;
  int np=16, val1_flag, f1, f2;
  int temp, pflag[np], i, *ptri, len[np], which, sum, nameLen;
  char *parms[16] = {"stepscale", "total_iteration", "warm", "savestep", "lowE", "maxE",\
                     "scale", "rho", "temperature", "maxvar", "response_col", "check_interactions",\
                     "Model_maxn", "seed", "gamma", "maxbeta"};

  char **ptrc, name[MAXFILELEN], c1;

  for (i=0, ptri=pflag; i<np; i++, ptri++) *ptri = 0;
  for (i=0, ptri=len, ptrc=parms; i<np; i++, ptri++, ptrc++) *ptri = (int) strlen(*ptrc); 

  fpt = open_file(parmfile, "r");
  
  /* Determine the order from row 1*/
  temp = fscanf(fpt, "%s %s", value, name);
  f1 = strncmp("VALUE", value, 5);
  f2 = strncmp("VALUE", name, 5);
  if ((!f1) && (!f2)) {
    printf("\n ERROR: Row 1 of input file must have the column names VALUE and NAME \n");
    fclose(fpt);
    exit(1);
  }
  if (!f1) {
    val1_flag = 1;
  } else {
    val1_flag = 0;
  }

  while(!feof(fpt)) 
  {   
    if (val1_flag) {
      temp = fscanf(fpt, "%s %s", value, name);
    } else {
      temp = fscanf(fpt, "%s %s", name, value);
    }
    if (temp == EOF) break;

    which = -1;
    nameLen = (int) strlen(name);
    c1 = name[0];
    for (i=0, ptrc=parms, ptri=len; i<np; i++, ptrc++, ptri++) {
      if ((nameLen == *ptri) && (c1 == **ptrc)) {
        temp = strncmp(name, *ptrc, *ptri);
        if (!temp) {
          which = i; 
          break;
        }
      }
    }

    if (which == -1) continue;
    pflag[which] = 1;
    x = atof(value);

    switch (which) {
      case 0:
        datastr->stepscale = (int) x;
        break;
      case 1:
        datastr->total_iteration = (int) x;
        break;
      case 2:
        datastr->warm = (int) x;
        break;
      case 3:
        datastr->savestep = (int) x;
        break;
      case 4:
        datastr->lowE = x;
        break;
      case 5:
        datastr->maxE = x;
        break;
      case 6:
        datastr->scale = x;
        break;
      case 7:
        datastr->rho = x;
        break;
      case 8:
        datastr->temperature = x;
        break;
      case 9:
        datastr->maxvar = (int) x;
        break;
      case 10:
        datastr->response_col = (int) x;
        break;
      case 11:
        datastr->check_interactions = (int) x;
        break;
      case 12:
        datastr->Model_maxn = (int) x;
        break;
      case 13:
        datastr->seed = (int) x;
        break;
      case 14:
        datastr->gamma = x;
        break;
      case 15:
        datastr->maxbeta = x;
        break;
    }
  }
  fclose(fpt);
  
  /* Check that all parms were specified */
  sum = 0;
  pflag[12] = 1;
  pflag[13] = 1;
  pflag[15] = 1;
  for (i=0, ptri=pflag; i<np; i++, ptri++) sum += *ptri; 
  if (sum < np) {
    printf("\n ERROR: The following (case sensitive) parameters were not specified: \n");
    for (i=0, ptri=pflag, ptrc=parms; i<np; i++, ptri++, ptrc++) {
      if (!*ptri) printf("%s \n", *ptrc);
    }
    exit(1);
  }
} 

/* Function to set up and return the hist matrix */
double ** set_hist(parmfile, grid, histFlag, parms)
char *parmfile;
int *grid, *histFlag;
PARMS *parms;
{
  char *tempstr;
  FILE *fpt;
  int i, j, n;
  double **hist=NULL;
  float x1, x2, x3, x4;
  double lowE=parms->lowE, scale=parms->scale;

  tempstr = (char *) malloc(MAXFILELEN*sizeof(char));
  CHECK_MEM(tempstr);
  
  i = get_option(parmfile, "in_hist", tempstr, 0);
  if (i) {
    /* Input file not specified */
    free(tempstr);
    if (*grid < 1) {
      printf("ERROR: Number of rows in hist matrix is less than 1. Reset maxE and/or lowE. \n");
      /*exit(0);*/
      *histFlag = 1;
      return(NULL);
    }

    hist = dmatrix(0, *grid, 1, 4);
    
    
    for(j=0; j<=*grid; j++){
        hist[j][1]=lowE+j*1.0/scale; 
       /* this is the upper bound for that interval */
	hist[j][2]=0.0;
	hist[j][3]=0.0;
        hist[j][4]=0.0;
    }

    *histFlag = 0;
    return(hist);
  }

  /* Input file was specified */
  fpt = open_file(tempstr, "r");
  free(tempstr);
  *histFlag = 1;

  /* Get the number of rows */
  fscanf(fpt, "%d", &n);
  if (n < 2) {
    printf("ERROR: Number of rows in file in_hist is less than 1\n");
    *histFlag = 2;
    fclose(fpt);
    return(hist);
    /*exit(0);*/
  }

  hist = dmatrix(0, n-1, 1, 4);

  /* 
  if (*grid != (n-1))
    {
       printf("ERROR: Number of rows in file in_hist is not correct \n");
       *histFlag = 3;
       fclose(fpt);
       return(hist);
    }
  */

  *grid = n-1;

  /* Read in each row */
  for (i=0; i<=n-1; i++) {
    fscanf(fpt, "%g %g %g %g", &x1, &x2, &x3, &x4);
    hist[i][1] = x1;
    hist[i][2] = x2;
    hist[i][3] = x3;
    hist[i][4] = x4;
  }
  fclose(fpt);

  *histFlag = 0;
  return(hist);

} /* END: set_hist */



/************************************************************************/
void print_dVec(vec, n, name)
double *vec;
int n;
char name[10];
{
  int i;
  printf("%s \n", name);
  for (i=1; i<=n; i++) {
    printf(" %g ", vec[i]);
  }
  printf("\n");
}

void print_dMat(mat, nr0, nr1, nc0, nc1, name)
double **mat;
int nr0, nr1, nc0, nc1;
char name[10];
{
  int i, j;
  printf("%s \n", name);
  for (i=nr0; i<=nr1; i++) {
    for (j=nc0; j<=nc1; j++) printf(" %g ", mat[i][j]);
    printf("\n");
  }
  printf("\n \n");
}

void print_iMat(mat, nr, nc, name)
int **mat;
int nr, nc;
char name[10];
{
  int i, j;
  printf("%s \n", name);
  for (i=1; i<=nr; i++) {
    for (j=1; j<=nc; j++) printf(" %d ", mat[i][j]);
    printf("\n");
  }
  printf("\n \n");
}

void print_iVec(vec, n, name)
int *vec;
int n;
char name[10];
{
  int i;
  printf("%s \n", name);
  for (i=1; i<=n; i++) {
    printf(" %d ", vec[i]);
  }
  printf("\n");
}

/************************************************************************************************/
/**************** For Models ********************************************************************/
/************************************************************************************************/

/* Function to compare vector vec to vector base. Returns 1 if vec > base, -1 if vec < base, or 0 if vec = base. */
static int compareVecs(base, m, vec, n)
int *base, *vec, m, n;
{
  int i, id0, id1, min_nm;

  if ((!m) && (!n)) {
    /* Both models have no groups (variables) */
    return(0);
  } else if (!m) {
    /* Base model has no groups (first model) */
    return(1);
  } else if (!n) {
    /* Model defined by vec has no vars */
    return(-1);
  }
  min_nm = MIN(n, m);

  for (i=1; i<=min_nm; i++) {
    id0 = base[i];
    id1 = vec[i];
    if (id1 > id0) {
      return(1);
    } else if (id1 < id0) {
      return(-1);
    } 
  }

  /* At this point, either vecs are equal or one is longer than the other */
  if (n == m) {
    return(0);
  } else if (n > m) {
    return(1);
  } else {
    return(-1);
  }

} /* END: compareVecs */

/* Function to check whether the vector of group numbers corresponds to a model that was previuously run. 
   If not previously run, then set insertRow to the row number where the model should go.
   Assume the first index in the vector is 1 not 0.
   Use a binary search method. Returns 0 for no match, or the row number in Models if a match is found. */
int find_row(vec, n, insertRow, parms)
int *vec;    /* vector of group numbers (in increasing order) to search for in models */
int n;      /* Number of elements of vec */
int *insertRow; /* Row number to insert vec into the matrix when the return value is 0 (no match).
                   All rows on or below insertRow should be shifted down 1 */
PARMS *parms;
{
  int i, val, mid, mid0, lower, upper, Model_n=parms->Model_n;
  RET_COST *ptrm, **Models=parms->Models;

  *insertRow = 0;

  /* If no models, then this one will be the first */
  if (!Model_n) {
    *insertRow = 1;
    return(0);
  }

  /* Check the endpoints */
  ptrm = Models[1];
  val = compareVecs(ptrm->groups, ptrm->num_grp, vec, n);
  if (!val) {
    /* A match */
    return(1);
  } else if (val < 0) {
    /* Outside the range of models, no possible match */
    *insertRow = 1;
    return(0);
  }

  ptrm = Models[Model_n];
  val = compareVecs(ptrm->groups, ptrm->num_grp, vec, n);
  if (!val) {
    /* A match */
    return(Model_n);
  } else if (val > 0) {
    /* Outside the range of models, no possible match */
    *insertRow = Model_n + 1;
    return(0);
  }

  mid0  = 0;
  lower = 1;
  upper = Model_n;

  /* Binary search */
  for (i=1; i<=Model_n; i++) {
    mid = (lower + upper)/2;
    if (mid == mid0) {
      *insertRow = upper;
      return(0);
    }
    
    ptrm = Models[mid];
    val = compareVecs(ptrm->groups, ptrm->num_grp, vec, n);
    if (val > 0) {
      /* Update the lower endpoint */
      lower = mid;
    } else if (val < 0) {
      /* Update the upper endpoint */
      upper = mid;
    } else if (!val) {
      /* A match */
      return(mid);
    } 
    mid0 = mid;
  }	

  return(0);

} /* END: find_row */

/* Function to insert a new model into the Models array and returns the updated number of models */
void insert_model(insertRow, rcost, parms)
int insertRow;
RET_COST *rcost;
PARMS *parms;
{
  RET_COST *new;
  int i;
  int Model_n=parms->Model_n, Model_maxn=parms->Model_maxn;
  RET_COST **Models=parms->Models;
/*
printf("insertRow=%d, Model_maxn=%d, Model_n=%d\n", insertRow, Model_maxn, Model_n);
*/
  if (Model_n >= Model_maxn) {
    /* Do not add any new models 
    printf("WARNING: Model could not be added. Consider increasing the parameter model_maxn\n");
    */

    return;

    /* Allocate more memory */
    /*
    printf("model_maxn was too small, allocating more memory.\n");
    saveModels = Models;
    newN = Model_maxn + MODEL_ADDN;
    Models = (RET_COST **) malloc((newN+1)*sizeof(RET_COST *));
    CHECK_MEM(Models);
    Model_maxn = newN;
    for (i=0; i<=Model_n; i++) Models[i] = saveModels[i];
    free(saveModels);
    */
   
  }
  if (insertRow < 1) {
    printf("ERROR: with insertRow\n");
    exit(0);
  }

  /* Allocate memory for a new model, and copy rcost except Yhat */ 
  new = cost_create(rcost, 0);

  /* Update the Models global variables */
  if (insertRow == Model_n + 1) {
    Models[insertRow] = new;
  } else if (insertRow <= Model_n) {
    /* Move pointers on and after insertRow down by 1 */
    for (i=Model_n; i>=insertRow; i--) Models[i+1] = Models[i];
    
    Models[insertRow] = new;
  } else {
    printf("ERROR: with insertRow and/or Model_n \n");
    exit(0);
  }

  parms->Model_n++;

} /* END: insert_model */


/* Function to compute the posterior probabilities */
void post_prob(row, pvec, parms)
int row;       /* Row of global Models vector that is a match */
double *pvec;  /* Vector to contain the computed probabilities */
PARMS *parms;
{
  RET_COST *rcost;
  int j, k, nbeta, *col_ids, DataNum=parms->DataNum;
  double mu, *beta, temp, **DataX=parms->DataX;

  rcost   = parms->Models[row];
  nbeta   = rcost->dim;
  beta    = rcost->beta;  
  col_ids = rcost->col_ids;

  for(k=1; k<=DataNum; k++){
    mu = 0.0;
    for (j=1; j<=nbeta; j++) mu += DataX[k][col_ids[j]]*beta[j];
    
    /* For predicted probabilities */
    temp = exp(mu);
    pvec[k] = temp/(1.0 + temp);    
  }

} /* END: post_prob */

/* Function to return the model with the smallest value */
int model_max_value(parms)
PARMS *parms;
{
  int i, mini=-1;
  double temp, minval;
  RET_COST **Models=parms->Models;

  minval = 1.0e300;
  for (i=1; i<=parms->Model_n; i++) {
    if (Models[i]->converged > 0) { 
      temp = Models[i]->value;
      if (temp < minval) {
        minval = temp;
        mini   = i;
      }
    }
  }
  if (mini == -1) {
    printf("ERROR: No model has converged\n");
    exit(0);
  }

  return(mini);

} /* END: model_max_value */

/* Function to copy an integer vector */
void copy_ivec(v1, v2, start, stop)
int *v1, *v2, start, stop;
{
  int i;

  if (stop >= start) {
    for (i=start; i<=stop; i++) v1[i] = v2[i];
  }

} /* copy_ivec */

/* Function to copy a double vector */
void copy_dvec(v1, v2, start, stop)
double *v1, *v2;
int start, stop;
{
  int i;

  if (stop >= start) {
    for (i=start; i<=stop; i++) v1[i] = v2[i];
  }

} /* copy_dvec */

/* Function to free a matrix */
void matrix_free(x, n)
void **x;
int n;
{
  int i;
  for (i=0; i<n; i++) {
    if (x[i] != NULL) free(x[i]);
  }
  free(x);

} /* END: matrix_free */
